# This module was inspired by MySQLPropPlugin

from sets import Set
import types

from BTrees.OOBTree import OOBTree
from Globals import DTMLFile, InitializeClass
from OFS.Folder import Folder
from Persistence import Persistent
from DateTime.DateTime import DateTime

from Products.CMFCore.utils import getToolByName

from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PluggableAuthService.utils import classImplements
from Products.PluggableAuthService.interfaces.plugins import IPropertiesPlugin
from Products.PluggableAuthService.interfaces.plugins import IUpdatePlugin
from Products.PluggableAuthService.UserPropertySheet import _guessSchema

from Products.ZSQLMethods.SQL import SQL

from Products.SQLPASPlugin import config as sqlpasconfig
from Products.SQLPASPlugin.utils import log, logEx

from Products.SQLPASPlugin.plugins.sql import SQLBase

# TODO: fix the dependency on PlonePAS
from Products.PlonePAS.interfaces.plugins import IMutablePropertiesPlugin
from Products.PlonePAS.sheet import MutablePropertySheet


def manage_addSQLPropertyProvider(self, id, title='', sql_connection='',
                                  RESPONSE=None, schema=None, **kw):
    """
    Create an instance of a mutable property manager.
    """
    o = SQLPropertyProvider(id, title, sql_connection, schema, **kw)
    self._setObject(o.getId(), o)

    if RESPONSE is not None:
        RESPONSE.redirect('manage_workspace')

manage_addSQLPropertyProviderForm = DTMLFile(
    "../zmi/MutablePropertyProviderForm", globals())


def _findAction(l, action):
    x = 0
    found = None
    while x < len(l) and found is None:
        if l[x]['action'] == action:
            found = l[x]
        x += 1
        
    return found

class SQLPropertyProvider(SQLBase):
    """Storage for mutable properties in SQL for users/groups.
    """

    meta_type = 'SQL Mutable Property Provider'
    
    
    def __init__(self, id, title='', connection='', schema=None, **kw):
        """Create SQL mutable property provider.

        Provide a schema either as a list of (name,type,value) tuples
        in the 'schema' parameter or as a series of keyword parameters
        'name=value'. Types will be guessed in this case.

        The 'value' is meant as the default value, and will be used
        unless the user provides data.

        If no schema is provided by constructor, the properties of the
        portal_memberdata object will be used.

        Types available: string, text, boolean, int, long, float, lines, date
        """
        self.id = id
        self.title = title
        
        # Probably need to setup the plugin with the sql connection info here
        # Need to do it as just a connection URL to keep it simple

        # calculate schema and default values
        defaultvalues = {}
        if not schema and not kw:
            schema = ()
        elif not schema and kw:
            schema = _guessSchema(kw)
            defaultvalues = kw
        else:
            schema = [(name, type) for name, type, value in schema]
            valuetuples = [(name, value) for name, type, value in schema]
            for name, value in valuetuples: defaultvalues[name] = value
        self._schema = tuple(schema)
        self._defaultvalues = defaultvalues

        SQLBase.__init__(self, id, title, connection)

        

    def _getSchema(self, isgroup=None):
        
        # this could probably stand to be cached
        datatool = isgroup and "portal_groupdata" or "portal_memberdata"

        schema = self._schema
        if not schema:
            # if no schema is provided, use portal_memberdata properties
            schema = ()
            #mdtool = getToolByName(self, datatool)
            #mdschema = mdtool.propertyMap()
            #schema = [(elt['id'], elt['type']) for elt in mdschema]
        return schema

    def _getDefaultValues(self, isgroup=None):
        """Returns a dictionary mapping of property names to default values.
        Defaults to portal_*data tool if necessary.
        """
        datatool = isgroup and "portal_groupdata" or "portal_memberdata"

        defaultvalues = self._defaultvalues
        if not self._schema:
            # if no schema is provided, use portal_*data properties
            defaultvalues = {}
            #mdtool = getToolByName(self, datatool)
            # we rely on propertyMap and propertyItems mapping
            #mdvalues = mdtool.propertyItems()
            #for name, value in mdvalues:
            #    defaultvalues[name] = value
        return defaultvalues

    def getPropertiesForUser(self, user, request=None):
        """Get property values for a user or group.
        Returns a dictionary of values or a PropertySheet.
        """
        
        isGroup = getattr(user, 'isGroup', lambda: None)()
        try:
            res = self.sqlLoadProperties(username=user.getUserName())
            all = res.dictionaries()
            if len(all) > 0:
                data = all[0]
                #Remove 'id' column if it exists in database result set
                if data.get('id', None):
                    data.pop('id')
            else:
                data = {}
        except:
            data = {}
            logEx("Error while trying to load properties for user")
        
        items = data.items()
        for key, value in items:
            _type = type(value)
            if value is None:
                value = ''
                data[key] = value
            elif str(_type).lower().startswith("<type 'date"):
                value = DateTime(str(value))
                data[key] = value
                                
        s = ''
        for key, value in data.items():
            s += 'key=%s; value=%s; type=%s\n' % (key, value, str(type(value)))
        
        data = self._remapNames(data)
        
        defaults = self._getDefaultValues(isGroup)

        # Convert from the generic db representation to the correct type
        for name, property_type in self._getSchema(isGroup) or ():
            if name in data.keys():
                if property_type == 'string' or property_type == 'text':
                    continue
                elif property_type == 'int':
                    data[name] = int(data[name])
                elif property_type == 'float':
                    data[name] = float(data[name])
                elif property_type == 'long':
                    data[name] = long(data[name])
                elif property_type == 'boolean':
                    if data[name] == '1':
                        data[name] = 1
                    else:
                        data[name] = 0
                elif property_type == 'date':
                    data[name] = DateTime(data[name])
                elif property_type == 'lines':
                    data[name] = tuple(data[name])
                else:
                    raise ValueError, 'Property %s: unknown type' % property_type            

        # provide default values where missing
        if not data: data = {}
        for key, val in defaults.items():
            if not data.has_key(key):
                data[key] = val

        # TODO: replace this hack to produce a fullname property with
        # a configurable piece that can produce virtual fields or something
        # similar
        
        if not data.get('fullname', None):
            fullname = data.get('firstname', '') + ' ' + data.get('lastname', '')
            fullname = fullname.strip()
            data['fullname'] = fullname
            
            data['_sql_special'] = ['fullname']

        sheet = MutablePropertySheet(self.id, **data)
        return sheet
        #return data
    
    
    def setPropertiesForUser(self, user, propertysheet):
        """Set the properties of a user or group based on the contents of a
        property sheet.
        """
        isGroup = getattr(user, 'isGroup', lambda: None)()
    
        props = dict(propertysheet.propertyItems())

        if props.has_key('_sql_special'):
            special = props['_sql_special']
            for x in special:
                del props[x]
            del props['_sql_special']


        self.updateUserInfo(user, set_id=None, set_info=props)
    
    
    def _findMapping(self, key):
        found = None
        prop = self.getProperty('col_mapping')
        x = 0
        while not found and x < len(prop):
            line = prop[x]
            if line.startswith(key+'/'):
                found = line[len(key+'/'):]
            
            x += 1
        
        return found
    
    def _findReverseMapping(self, key):
        found = None
        prop = self.getProperty('col_mapping')
        x = 0
        while not found and x < len(prop):
            line = prop[x]
            if line.endswith('/'+key):
                found = line[:line.find('/')]
            
            x += 1
        
        return found

    def _remapNames(self, data, reverse=False):
        res = {}
        for key, value in data.items():
            if not reverse:
                mapping = self._findMapping(key)
            else:
                mapping = self._findReverseMapping(key)
            k = key
            if mapping:
                k = mapping
            res[k] = value

        return res

    def updateUserInfo(self, user, set_id, set_info):
        if set_id is not None:
            raise NotImplementedError, "Cannot currently rename the user id of a user"
        
        users_table = self.getProperty('users_table')
        users_col_username = self.getProperty('users_col_username')

        realFields = self._remapNames(set_info, True)        

        sql = 'UPDATE %s SET ' % users_table
        for key, value in realFields.items():
            _type = 'string'
            if type(value) == types.FloatType:
                _type = 'float'
            elif type(value) == types.IntType:
                _type = 'int'
                
            sql += '"' + key + '"=<dtml-sqlvar %s type=%s>,' % (key, _type)
        sql = sql[:-1]
        sql += " WHERE %s='%s'" % (users_col_username, user.getUserName())
            
        params = ' '.join(realFields.keys())
        sqlMethod = SQL('random', 'random', self._connection, params, sql)
        sqlMethod = sqlMethod.__of__(self)
        
        data = dict(realFields)
        data[users_col_username] = user.getUserName()

        sqlMethod(**data)
        

    def __setstate__(self, state):
        Persistent.__setstate__(self, state)
        self.addSQLQueries()
        
    def getSQLQueries(self):
        return _SQL_QUERIES


classImplements(SQLPropertyProvider,
                IPropertiesPlugin, 
                IUpdatePlugin,
                IMutablePropertiesPlugin)

InitializeClass(SQLPropertyProvider)

_SQL_QUERIES = (
    (
     "sqlLoadProperties", 
     "Load member properties", 
     "username", 
     """
SELECT *
FROM %(users_table)s
WHERE %(username_col)s=<dtml-sqlvar username type=string>
     """
    ),

)
